@@ -22,6 +22,7 @@
 #include "common/version.h"
 #include "include/str_list.h"
 #include "include/types.h"
+#include "include/stringify.h"
 #include "msg/msg_types.h"
 #include "osd/osd_types.h"
 
@@ -62,7 +63,7 @@ struct ceph_file_layout g_default_file_layout = {
  fl_object_size: init_le32(1<<22),
  fl_cas_hash: init_le32(0),
  fl_object_stripe_unit: init_le32(0),
- fl_pg_preferred : init_le32(-1),
+ fl_unused: init_le32(-1),
  fl_pg_pool : init_le32(-1),
 };
 
@@ -174,7 +175,9 @@ void md_config_t::remove_observer(md_config_obs_t* observer_)
 }
 
 int md_config_t::parse_config_files(const char *conf_files,
-				    std::deque<std::string> *parse_errors, int flags)
+				    std::deque<std::string> *parse_errors,
+				    std::ostream *warnings,
+				    int flags)
 {
   Mutex::Locker l(lock);
   if (internal_safe_to_start_threads)
@@ -192,20 +195,22 @@ int md_config_t::parse_config_files(const char *conf_files,
   }
   std::list<std::string> cfl;
   get_str_list(conf_files, cfl);
-  return parse_config_files_impl(cfl, parse_errors);
+  return parse_config_files_impl(cfl, parse_errors, warnings);
 }
 
 int md_config_t::parse_config_files_impl(const std::list<std::string> &conf_files,
-					 std::deque<std::string> *parse_errors)
+					 std::deque<std::string> *parse_errors,
+					 std::ostream *warnings)
 {
-  Mutex::Locker l(lock);
+  assert(lock.is_locked());
+
   // open new conf
   list<string>::const_iterator c;
   for (c = conf_files.begin(); c != conf_files.end(); ++c) {
     cf.clear();
     string fn = *c;
     expand_meta(fn);
-    int ret = cf.parse_file(fn.c_str(), parse_errors);
+    int ret = cf.parse_file(fn.c_str(), parse_errors, warnings);
     if (ret == 0)
       break;
     else if (ret != -ENOENT)
@@ -215,11 +220,11 @@ int md_config_t::parse_config_files_impl(const std::list<std::string> &conf_file
     return -EINVAL;
 
   std::vector <std::string> my_sections;
-  get_my_sections(my_sections);
+  _get_my_sections(my_sections);
   for (int i = 0; i < NUM_CONFIG_OPTIONS; i++) {
     config_option *opt = &config_optionsp[i];
     std::string val;
-    int ret = get_val_from_conf_file(my_sections, opt->name, val, false);
+    int ret = _get_val_from_conf_file(my_sections, opt->name, val, false);
     if (ret == 0) {
       set_val_impl(val.c_str(), opt);
     }
@@ -230,7 +235,7 @@ int md_config_t::parse_config_files_impl(const std::list<std::string> &conf_file
     std::string as_option("debug_");
     as_option += subsys.get_name(o);
     std::string val;
-    int ret = get_val_from_conf_file(my_sections, as_option.c_str(), val, false);
+    int ret = _get_val_from_conf_file(my_sections, as_option.c_str(), val, false);
     if (ret == 0) {
       int log, gather;
       int r = sscanf(val.c_str(), "%d/%d", &log, &gather);
@@ -281,18 +286,46 @@ void md_config_t::parse_env()
 
 void md_config_t::show_config(std::ostream& out)
 {
-  out << "name = " << name << std::endl;
-  out << "cluster = " << cluster << std::endl;
+  Mutex::Locker l(lock);
+  _show_config(&out, NULL);
+}
+
+void md_config_t::show_config(Formatter *f)
+{
+  Mutex::Locker l(lock);
+  _show_config(NULL, f);
+}
+
+void md_config_t::_show_config(std::ostream *out, Formatter *f)
+{
+  if (out) {
+    *out << "name = " << name << std::endl;
+    *out << "cluster = " << cluster << std::endl;
+  }
+  if (f) {
+    f->dump_string("name", stringify(name));
+    f->dump_string("cluster", cluster);
+  }
   for (int o = 0; o < subsys.get_num(); o++) {
-    out << "debug_" << subsys.get_name(o)
-	<< " = " << subsys.get_log_level(o)
-	<< "/" << subsys.get_gather_level(o) << std::endl;
+    if (out)
+      *out << "debug_" << subsys.get_name(o)
+	   << " = " << subsys.get_log_level(o)
+	   << "/" << subsys.get_gather_level(o) << std::endl;
+    if (f) {
+      ostringstream ss;
+      ss << subsys.get_log_level(o)
+	 << "/" << subsys.get_gather_level(o);
+      f->dump_string(subsys.get_name(o).c_str(), ss.str());
+    }
   }
   for (int i = 0; i < NUM_CONFIG_OPTIONS; i++) {
     config_option *opt = config_optionsp + i;
     char *buf;
     _get_val(opt->name, &buf, -1);
-    out << opt->name << " = " << buf << std::endl;
+    if (out)
+      *out << opt->name << " = " << buf << std::endl;
+    if (f)
+      f->dump_string(opt->name, buf);
   }
 }
 
@@ -320,7 +353,7 @@ int md_config_t::parse_argv(std::vector<const char*>& args)
     }
     else if (ceph_argparse_flag(args, i, "--show_config", (char*)NULL)) {
       expand_all_meta();
-      show_config(cout);
+      _show_config(&cout, NULL);
       _exit(0);
     }
     else if (ceph_argparse_witharg(args, i, &val, "--show_config_value", (char*)NULL)) {
@@ -427,7 +460,7 @@ int md_config_t::parse_option(std::vector<const char*>& args,
       } else {
 	std::string no("--no-");
 	no += opt->name;
-	if (ceph_argparse_flag(args, i, &res, no.c_str(), (char*)NULL)) {
+	if (ceph_argparse_flag(args, i, no.c_str(), (char*)NULL)) {
 	  set_val_impl("false", opt);
 	  break;
 	}
@@ -578,7 +611,7 @@ void md_config_t::set_val_or_die(const char *key, const char *val)
   assert(ret == 0);
 }
 
-int md_config_t::set_val(const char *key, const char *val)
+int md_config_t::set_val(const char *key, const char *val, bool meta)
 {
   Mutex::Locker l(lock);
   if (!key)
@@ -587,7 +620,8 @@ int md_config_t::set_val(const char *key, const char *val)
     return -EINVAL;
 
   std::string v(val);
-  expand_meta(v);
+  if (meta)
+    expand_meta(v);
 
   string k(ConfFile::normalize_key_name(key));
 
@@ -695,6 +729,8 @@ int md_config_t::_get_val(const char *key, char **buf, int len) const
     int l = strlen(str.c_str()) + 1;
     if (len == -1) {
       *buf = (char*)malloc(l);
+      if (!*buf)
+        return -ENOMEM;
       strcpy(*buf, str.c_str());
       return 0;
     }
@@ -727,6 +763,12 @@ int md_config_t::_get_val(const char *key, char **buf, int len) const
 void md_config_t::get_my_sections(std::vector <std::string> &sections) const
 {
   Mutex::Locker l(lock);
+  _get_my_sections(sections);
+}
+
+void md_config_t::_get_my_sections(std::vector <std::string> &sections) const
+{
+  assert(lock.is_locked());
   sections.push_back(name.to_str());
 
   sections.push_back(name.get_type_name());
@@ -749,6 +791,13 @@ int md_config_t::get_val_from_conf_file(const std::vector <std::string> &section
 		    const char *key, std::string &out, bool emeta) const
 {
   Mutex::Locker l(lock);
+  return _get_val_from_conf_file(sections, key, out, emeta);
+}
+
+int md_config_t::_get_val_from_conf_file(const std::vector <std::string> &sections,
+					 const char *key, std::string &out, bool emeta) const
+{
+  assert(lock.is_locked());
   std::vector <std::string>::const_iterator s = sections.begin();
   std::vector <std::string>::const_iterator s_end = sections.end();
   for (; s != s_end; ++s) {
@@ -866,49 +915,105 @@ void md_config_t::expand_all_meta()
   }
 }
 
-bool md_config_t::expand_meta(std::string &val) const
+bool md_config_t::expand_meta(std::string &origval) const
 {
   assert(lock.is_locked());
   bool found_meta = false;
+
+  set<string> resolved;
+
+  string val = origval;
+
+ restart:
   string out;
-  string::size_type sz = val.size();
-  out.reserve(sz);
-  for (string::size_type s = 0; s < sz; ) {
+  out.reserve(val.size());
+
+  for (string::size_type s = 0; s < val.size(); ) {
     if (val[s] != '$') {
       out += val[s++];
       continue;
     }
-    string::size_type rem = sz - (s + 1);
-    int i;
-    for (i = 0; i < NUM_CONF_METAVARIABLES; ++i) {
-      size_t clen = strlen(CONF_METAVARIABLES[i]);
-      if (rem < clen)
-	continue;
-      if (strncmp(val.c_str() + s + 1, CONF_METAVARIABLES[i], clen))
-	continue;
-      if (strcmp(CONF_METAVARIABLES[i], "type")==0)
-	out += name.get_type_name();
-      else if (strcmp(CONF_METAVARIABLES[i], "cluster")==0)
-	out += cluster;
-      else if (strcmp(CONF_METAVARIABLES[i], "name")==0)
-	out += name.to_cstr();
-      else if (strcmp(CONF_METAVARIABLES[i], "host")==0)
-	out += host;
-      else if (strcmp(CONF_METAVARIABLES[i], "num")==0)
-	out += name.get_id().c_str();
-      else if (strcmp(CONF_METAVARIABLES[i], "id")==0)
-	out += name.get_id().c_str();
+
+    // try to parse the variable name into var, either \$\{(.+)\} or
+    // \$([a-z\_]+)
+    const char *valid_chars = "abcdefghijklmnopqrstuvwxyz_";
+    string var;
+    size_t endpos = 0;
+    if (val[s+1] == '{') {
+      // ...${foo_bar}...
+      endpos = val.find_first_not_of(valid_chars, s+2);
+      if (endpos != std::string::npos &&
+	  val[endpos] == '}') {
+	var = val.substr(s+2, endpos-s-2);
+	endpos++;
+      }
+    } else {
+      // ...$foo...
+      endpos = val.find_first_not_of(valid_chars, s+1);
+      if (endpos != std::string::npos)
+	var = val.substr(s+1, endpos-s-1);
       else
-	assert(0); // unreachable
-      found_meta = true;
-      break;
+	var = val.substr(s+1);
     }
-    if (i == NUM_CONF_METAVARIABLES)
-      out += val[s++];
-    else
-      s += strlen(CONF_METAVARIABLES[i]) + 1;
+    //cout << "var='" << var << "'" << std::endl;
+
+    if (var.length()) {
+      // special metavariable?
+      for (int i = 0; i < NUM_CONF_METAVARIABLES; ++i) {
+	if (var != CONF_METAVARIABLES[i])
+	  continue;
+	//cout << "  meta match of " << var << " " << CONF_METAVARIABLES[i] << std::endl;
+	if (var == "type")
+	  out += name.get_type_name();
+	else if (var == "cluster")
+	  out += cluster;
+	else if (var == "name")
+	  out += name.to_cstr();
+	else if (var == "host")
+	  out += host;
+	else if (var == "num")
+	  out += name.get_id().c_str();
+	else if (var == "id")
+	  out += name.get_id().c_str();
+	else
+	  assert(0); // unreachable
+	found_meta = true;
+	if (endpos != std::string::npos)
+	  out += val.substr(endpos);
+	//cout << "val '" << val << "' s " << s << " out '" << out << "'"  << std::endl;
+	val = out;
+	goto restart;
+      }
+
+      // config option?
+      for (int i = 0; i < NUM_CONFIG_OPTIONS; i++) {
+	config_option *opt = &config_optionsp[i];
+	if (var != opt->name)
+	  continue;
+	
+	// avoid loops
+	if (resolved.count(opt->name))
+	  continue;	// loop; skip
+	resolved.insert(opt->name);
+
+	found_meta = true;
+	char *vv = NULL;
+	_get_val(opt->name, &vv, -1);
+	out += vv;
+	if (endpos != std::string::npos)
+	  out += val.substr(endpos);
+	//cout << "val '" << val << "' s " << s << " out '" << out << "' after sub " << opt->name << " -> " << vv << std::endl;
+	val = out;
+	free(vv);
+	goto restart;
+      }
+    }
+
+    // pass it thru
+    out += val[s++];
   }
-  val = out;
+  //cout << "done '" << origval << "' -> '" << out << "'" << std::endl;
+  origval = out;
   return found_meta;
 }
 
