@@ -246,44 +246,49 @@ void Objecter::shutdown()
   }
 }
 
-void Objecter::send_linger(LingerOp *info, bool first_send)
+void Objecter::send_linger(LingerOp *info)
 {
-  if (!info->registering) {
-    ldout(cct, 15) << "send_linger " << info->linger_id << dendl;
-    vector<OSDOp> ops = info->ops; // need to pass a copy to ops
-    Context *onack = (!info->registered && info->on_reg_ack) ? new C_Linger_Ack(this, info) : NULL;
-    Context *oncommit = new C_Linger_Commit(this, info);
-    Op *o = new Op(info->oid, info->oloc, ops, info->flags | CEPH_OSD_FLAG_READ,
-		   onack, oncommit,
-		   info->pobjver);
-    o->snapid = info->snap;
+  ldout(cct, 15) << "send_linger " << info->linger_id << dendl;
+  vector<OSDOp> opv = info->ops; // need to pass a copy to ops
+  Context *onack = (!info->registered && info->on_reg_ack) ? new C_Linger_Ack(this, info) : NULL;
+  Context *oncommit = new C_Linger_Commit(this, info);
+  Op *o = new Op(info->oid, info->oloc, opv, info->flags | CEPH_OSD_FLAG_READ,
+		 onack, oncommit,
+		 info->pobjver);
+  o->snapid = info->snap;
 
-    if (info->session) {
-      int r = recalc_op_target(o);
-      if (r == RECALC_OP_TARGET_POOL_DNE) {
-	linger_check_for_latest_map(info);
-      }
+  // do not resend this; we will send a new op to reregister
+  o->should_resend = false;
+
+  if (info->session) {
+    int r = recalc_op_target(o);
+    if (r == RECALC_OP_TARGET_POOL_DNE) {
+      linger_check_for_latest_map(info);
     }
-
-    if (first_send) {
-      op_submit(o);
-    } else {
-      _op_submit(o);
-    }
-
-    OSDSession *s = o->session;
-    if (info->session != s) {
-      info->session_item.remove_myself();
-      info->session = s;
-      if (info->session)
-	s->linger_ops.push_back(&info->session_item);
-    }
-    info->registering = true;
-
-    logger->inc(l_osdc_linger_send);
-  } else {
-    ldout(cct, 15) << "send_linger " << info->linger_id << " already (re)registering" << dendl;
   }
+
+  if (info->register_tid) {
+    // repeat send.  cancel old registeration op, if any.
+    if (ops.count(info->register_tid)) {
+      Op *o = ops[info->register_tid];
+      cancel_op(o);
+    }
+    info->register_tid = _op_submit(o);
+  } else {
+    // first send
+    info->register_tid = op_submit(o);
+  }
+
+  OSDSession *s = o->session;
+  if (info->session != s) {
+    info->session_item.remove_myself();
+    info->session = s;
+    if (info->session)
+      s->linger_ops.push_back(&info->session_item);
+  }
+  info->registering = true;
+
+  logger->inc(l_osdc_linger_send);
 }
 
 void Objecter::_linger_ack(LingerOp *info, int r) 
@@ -348,7 +353,7 @@ tid_t Objecter::linger(const object_t& oid, const object_locator_t& oloc,
 
   logger->set(l_osdc_linger_active, linger_ops.size());
 
-  send_linger(info, true);
+  send_linger(info);
 
   return info->linger_id;
 }
@@ -450,6 +455,7 @@ void Objecter::handle_osd_map(MOSDMap *m)
 	     p != linger_ops.end();
 	     p++) {
 	  LingerOp *op = p->second;
+	  ldout(cct, 10) << " checking linger op " << op->linger_id << dendl;
 	  int r = recalc_linger_op_target(op);
 	  if (skipped_map)
 	    r = RECALC_OP_TARGET_NEED_RESEND;
@@ -472,6 +478,7 @@ void Objecter::handle_osd_map(MOSDMap *m)
 	     p != ops.end();
 	     ++p) {
 	  Op *op = p->second;
+	  ldout(cct, 10) << " checking op " << op->tid << dendl;
 	  int r = recalc_op_target(op);
 	  if (skipped_map)
 	    r = RECALC_OP_TARGET_NEED_RESEND;
@@ -541,16 +548,20 @@ void Objecter::handle_osd_map(MOSDMap *m)
   // resend requests
   for (map<tid_t, Op*>::iterator p = need_resend.begin(); p != need_resend.end(); p++) {
     Op *op = p->second;
-    if (op->session) {
-      logger->inc(l_osdc_op_resend);
-      send_op(op);
+    if (op->should_resend) {
+      if (op->session) {
+	logger->inc(l_osdc_op_resend);
+	send_op(op);
+      }
+    } else {
+      cancel_op(op);
     }
   }
   for (list<LingerOp*>::iterator p = need_resend_linger.begin(); p != need_resend_linger.end(); p++) {
     LingerOp *op = *p;
     if (op->session) {
       logger->inc(l_osdc_linger_resend);
-      send_linger(op, false);
+      send_linger(op);
     }
   }
 
@@ -747,15 +758,31 @@ void Objecter::kick_requests(OSDSession *session)
   ldout(cct, 10) << "kick_requests for osd." << session->osd << dendl;
 
   // resend ops
-  for (xlist<Op*>::iterator p = session->ops.begin(); !p.end(); ++p) {
+  map<tid_t,Op*> resend;  // resend in tid order
+  for (xlist<Op*>::iterator p = session->ops.begin(); !p.end();) {
+    Op *op = *p;
+    ++p;
     logger->inc(l_osdc_op_resend);
-    send_op(*p);
+    if (op->should_resend) {
+      resend[op->tid] = op;
+    } else {
+      cancel_op(op);
+    }
+  }
+  while (!resend.empty()) {
+    send_op(resend.begin()->second);
+    resend.erase(resend.begin());
   }
 
   // resend lingers
+  map<uint64_t, LingerOp*> lresend;  // resend in order
   for (xlist<LingerOp*>::iterator j = session->linger_ops.begin(); !j.end(); ++j) {
     logger->inc(l_osdc_linger_resend);
-    send_linger(*j, false);
+    lresend[(*j)->linger_id] = *j;
+  }
+  while (!lresend.empty()) {
+    send_linger(lresend.begin()->second);
+    lresend.erase(lresend.begin());
   }
 }
 
@@ -826,29 +853,21 @@ void Objecter::tick()
 
 void Objecter::resend_mon_ops()
 {
-  utime_t cutoff = ceph_clock_now(cct);
-  cutoff -= cct->_conf->objecter_mon_retry_interval;
-
+  ldout(cct, 10) << "resend_mon_ops" << dendl;
 
   for (map<tid_t,PoolStatOp*>::iterator p = poolstat_ops.begin(); p!=poolstat_ops.end(); ++p) {
-    if (p->second->last_submit < cutoff) {
-      poolstat_submit(p->second);
-      logger->inc(l_osdc_poolstat_resend);
-    }
+    poolstat_submit(p->second);
+    logger->inc(l_osdc_poolstat_resend);
   }
 
   for (map<tid_t,StatfsOp*>::iterator p = statfs_ops.begin(); p!=statfs_ops.end(); ++p) {
-    if (p->second->last_submit < cutoff) {
-      fs_stats_submit(p->second);
-      logger->inc(l_osdc_statfs_resend);
-    }
+    fs_stats_submit(p->second);
+    logger->inc(l_osdc_statfs_resend);
   }
 
   for (map<tid_t,PoolOp*>::iterator p = pool_ops.begin(); p!=pool_ops.end(); ++p) {
-    if (p->second->last_submit < cutoff) {
-      pool_op_submit(p->second);
-      logger->inc(l_osdc_poolop_resend);
-    }
+    pool_op_submit(p->second);
+    logger->inc(l_osdc_poolop_resend);
   }
 }
 
@@ -1086,6 +1105,35 @@ bool Objecter::recalc_linger_op_target(LingerOp *linger_op)
   return RECALC_OP_TARGET_NO_ACTION;
 }
 
+void Objecter::cancel_op(Op *op)
+{
+  ldout(cct, 15) << "cancel_op " << op->tid << dendl;
+
+  // currently this only works for linger registrations, since we just
+  // throw out the callbacks.
+  assert(!op->should_resend);
+  delete op->onack;
+  delete op->oncommit;
+
+  finish_op(op);
+}
+
+void Objecter::finish_op(Op *op)
+{
+  ldout(cct, 15) << "finish_op " << op->tid << dendl;
+
+  op->session_item.remove_myself();
+  if (op->budgeted)
+    put_op_budget(op);
+  if (op->con)
+    op->con->put();
+
+  ops.erase(op->tid);
+  logger->set(l_osdc_op_active, ops.size());
+
+  delete op;
+}
+
 void Objecter::send_op(Op *op)
 {
   ldout(cct, 15) << "send_op " << op->tid << " to osd." << op->session->osd << dendl;
@@ -1292,15 +1340,8 @@ void Objecter::handle_osd_op_reply(MOSDOpReply *m)
 
   // done with this tid?
   if (!op->onack && !op->oncommit) {
-    op->session_item.remove_myself();
     ldout(cct, 15) << "handle_osd_op_reply completed tid " << tid << dendl;
-    if (op->budgeted)
-      put_op_budget(op);
-    ops.erase(tid);
-    logger->set(l_osdc_op_active, ops.size());
-    if (op->con)
-      op->con->put();
-    delete op;
+    finish_op(op);
   }
   
   ldout(cct, 5) << num_unacked << " unacked, " << num_uncommitted << " uncommitted" << dendl;
@@ -1449,14 +1490,22 @@ void Objecter::_list_reply(ListContext *list_context, int r, bufferlist *bl,
 
 //snapshots
 
-int Objecter::create_pool_snap(int64_t pool, string& snapName, Context *onfinish) {
-  ldout(cct, 10) << "create_pool_snap; pool: " << pool << "; snap: " << snapName << dendl;
+int Objecter::create_pool_snap(int64_t pool, string& snap_name, Context *onfinish)
+{
+  ldout(cct, 10) << "create_pool_snap; pool: " << pool << "; snap: " << snap_name << dendl;
+
+  const pg_pool_t *p = osdmap->get_pg_pool(pool);
+  if (!p)
+    return -EINVAL;
+  if (p->snap_exists(snap_name.c_str()))
+    return -EEXIST;
+
   PoolOp *op = new PoolOp;
   if (!op)
     return -ENOMEM;
   op->tid = ++last_tid;
   op->pool = pool;
-  op->name = snapName;
+  op->name = snap_name;
   op->onfinish = onfinish;
   op->pool_op = POOL_OP_CREATE_SNAP;
   pool_ops[op->tid] = op;
@@ -1499,15 +1548,22 @@ int Objecter::allocate_selfmanaged_snap(int64_t pool, snapid_t *psnapid,
   return 0;
 }
 
-int Objecter::delete_pool_snap(int64_t pool, string& snapName, Context *onfinish)
+int Objecter::delete_pool_snap(int64_t pool, string& snap_name, Context *onfinish)
 {
-  ldout(cct, 10) << "delete_pool_snap; pool: " << pool << "; snap: " << snapName << dendl;
+  ldout(cct, 10) << "delete_pool_snap; pool: " << pool << "; snap: " << snap_name << dendl;
+
+  const pg_pool_t *p = osdmap->get_pg_pool(pool);
+  if (!p)
+    return -EINVAL;
+  if (!p->snap_exists(snap_name.c_str()))
+    return -ENOENT;
+
   PoolOp *op = new PoolOp;
   if (!op)
     return -ENOMEM;
   op->tid = ++last_tid;
   op->pool = pool;
-  op->name = snapName;
+  op->name = snap_name;
   op->onfinish = onfinish;
   op->pool_op = POOL_OP_DELETE_SNAP;
   pool_ops[op->tid] = op;
@@ -1539,6 +1595,10 @@ int Objecter::create_pool(string& name, Context *onfinish, uint64_t auid,
 			  int crush_rule)
 {
   ldout(cct, 10) << "create_pool name=" << name << dendl;
+
+  if (osdmap->lookup_pg_pool_name(name.c_str()) >= 0)
+    return -EEXIST;
+
   PoolOp *op = new PoolOp;
   if (!op)
     return -ENOMEM;
@@ -1560,6 +1620,9 @@ int Objecter::delete_pool(int64_t pool, Context *onfinish)
 {
   ldout(cct, 10) << "delete_pool " << pool << dendl;
 
+  if (!osdmap->have_pg_pool(pool))
+    return -ENOENT;
+
   PoolOp *op = new PoolOp;
   if (!op) return -ENOMEM;
   op->tid = ++last_tid;
@@ -1879,6 +1942,7 @@ void Objecter::_sg_read_finish(vector<ObjectExtent>& extents, vector<bufferlist>
 
 void Objecter::ms_handle_connect(Connection *con)
 {
+  ldout(cct, 10) << "ms_handle_connect " << con << dendl;
   if (con->get_peer_type() == CEPH_ENTITY_TYPE_MON)
     resend_mon_ops();
 }
