@@ -6,6 +6,7 @@
 #include <map>
 #include "include/types.h"
 #include "include/utime.h"
+#include "include/assert.h"
 
 enum {
   UPDATE_OBJ,
@@ -17,6 +18,8 @@ enum {
 #define CACHE_FLAG_META           0x4
 #define CACHE_FLAG_MODIFY_XATTRS  0x8
 
+#define mydout(v) lsubdout(T::cct, rgw, v)
+
 struct ObjectMetaInfo {
   uint64_t size;
   time_t mtime;
@@ -182,7 +185,7 @@ class RGWCache  : public T
   void finalize() {
     T::finalize_watch();
   }
-  int distribute(rgw_obj& obj, ObjectCacheInfo& obj_info, int op);
+  int distribute(const string& normal_name, rgw_obj& obj, ObjectCacheInfo& obj_info, int op);
   int watch_cb(int opcode, uint64_t ver, bufferlist& bl);
 public:
   RGWCache() {}
@@ -193,7 +196,8 @@ public:
                 map<string, bufferlist>* rmattrs);
   int put_obj_meta(void *ctx, rgw_obj& obj, uint64_t size, time_t *mtime,
                    map<std::string, bufferlist>& attrs, RGWObjCategory category, bool exclusive,
-                   map<std::string, bufferlist>* rmattrs, const bufferlist *data);
+                   map<std::string, bufferlist>* rmattrs, const bufferlist *data,
+                   RGWObjManifest *manifest);
 
   int put_obj_data(void *ctx, rgw_obj& obj, const char *data,
               off_t ofs, size_t len, bool exclusive);
@@ -219,7 +223,7 @@ int RGWCache<T>::delete_obj(void *ctx, rgw_obj& obj, bool sync)
   cache.remove(name);
 
   ObjectCacheInfo info;
-  distribute(obj, info, REMOVE_OBJ);
+  distribute(name, obj, info, REMOVE_OBJ);
 
   return T::delete_obj(ctx, obj, sync);
 }
@@ -288,9 +292,9 @@ int RGWCache<T>::set_attr(void *ctx, rgw_obj& obj, const char *attr_name, buffer
     string name = normal_name(bucket, oid);
     if (ret >= 0) {
       cache.put(name, info);
-      int r = distribute(obj, info, UPDATE_OBJ);
+      int r = distribute(name, obj, info, UPDATE_OBJ);
       if (r < 0)
-        ldout(T::cct, 0) << "ERROR: failed to distribute cache for " << obj << dendl;
+        mydout(0) << "ERROR: failed to distribute cache for " << obj << dendl;
     } else {
      cache.remove(name);
     }
@@ -322,9 +326,9 @@ int RGWCache<T>::set_attrs(void *ctx, rgw_obj& obj,
     string name = normal_name(bucket, oid);
     if (ret >= 0) {
       cache.put(name, info);
-      int r = distribute(obj, info, UPDATE_OBJ);
+      int r = distribute(name, obj, info, UPDATE_OBJ);
       if (r < 0)
-        ldout(T::cct, 0) << "ERROR: failed to distribute cache for " << obj << dendl;
+        mydout(0) << "ERROR: failed to distribute cache for " << obj << dendl;
     } else {
      cache.remove(name);
     }
@@ -336,7 +340,7 @@ int RGWCache<T>::set_attrs(void *ctx, rgw_obj& obj,
 template <class T>
 int RGWCache<T>::put_obj_meta(void *ctx, rgw_obj& obj, uint64_t size, time_t *mtime,
                               map<std::string, bufferlist>& attrs, RGWObjCategory category, bool exclusive,
-                              map<std::string, bufferlist>* rmattrs, const bufferlist *data)
+                              map<std::string, bufferlist>* rmattrs, const bufferlist *data, RGWObjManifest *manifest)
 {
   rgw_bucket bucket;
   string oid;
@@ -353,14 +357,14 @@ int RGWCache<T>::put_obj_meta(void *ctx, rgw_obj& obj, uint64_t size, time_t *mt
       info.flags |= CACHE_FLAG_DATA;
     }
   }
-  int ret = T::put_obj_meta(ctx, obj, size, mtime, attrs, category, exclusive, rmattrs, data);
+  int ret = T::put_obj_meta(ctx, obj, size, mtime, attrs, category, exclusive, rmattrs, data, manifest);
   if (cacheable) {
     string name = normal_name(bucket, oid);
     if (ret >= 0) {
       cache.put(name, info);
-      int r = distribute(obj, info, UPDATE_OBJ);
+      int r = distribute(name, obj, info, UPDATE_OBJ);
       if (r < 0)
-        ldout(T::cct, 0) << "ERROR: failed to distribute cache for " << obj << dendl;
+        mydout(0) << "ERROR: failed to distribute cache for " << obj << dendl;
     } else {
      cache.remove(name);
     }
@@ -393,9 +397,9 @@ int RGWCache<T>::put_obj_data(void *ctx, rgw_obj& obj, const char *data,
     string name = normal_name(bucket, oid);
     if (ret >= 0) {
       cache.put(name, info);
-      int r = distribute(obj, info, UPDATE_OBJ);
+      int r = distribute(name, obj, info, UPDATE_OBJ);
       if (r < 0)
-        ldout(T::cct, 0) << "ERROR: failed to distribute cache for " << obj << dendl;
+        mydout(0) << "ERROR: failed to distribute cache for " << obj << dendl;
     } else {
      cache.remove(name);
     }
@@ -452,7 +456,7 @@ done:
 }
 
 template <class T>
-int RGWCache<T>::distribute(rgw_obj& obj, ObjectCacheInfo& obj_info, int op)
+int RGWCache<T>::distribute(const string& normal_name, rgw_obj& obj, ObjectCacheInfo& obj_info, int op)
 {
   RGWCacheNotifyInfo info;
 
@@ -462,7 +466,7 @@ int RGWCache<T>::distribute(rgw_obj& obj, ObjectCacheInfo& obj_info, int op)
   info.obj = obj;
   bufferlist bl;
   ::encode(info, bl);
-  int ret = T::distribute(bl);
+  int ret = T::distribute(normal_name, bl);
   return ret;
 }
 
@@ -475,14 +479,17 @@ int RGWCache<T>::watch_cb(int opcode, uint64_t ver, bufferlist& bl)
     bufferlist::iterator iter = bl.begin();
     ::decode(info, iter);
   } catch (buffer::end_of_buffer& err) {
-    ldout(T::cct, 0) << "ERROR: got bad notification" << dendl;
+    mydout(0) << "ERROR: got bad notification" << dendl;
     return -EIO;
   } catch (buffer::error& err) {
-    ldout(T::cct, 0) << "ERROR: buffer::error" << dendl;
+    mydout(0) << "ERROR: buffer::error" << dendl;
     return -EIO;
   }
 
-  string name = normal_name(info.obj);
+  rgw_bucket bucket;
+  string oid;
+  normalize_bucket_and_obj(info.obj.bucket, info.obj.object, bucket, oid);
+  string name = normal_name(bucket, oid);
 
   switch (info.op) {
   case UPDATE_OBJ:
@@ -492,7 +499,7 @@ int RGWCache<T>::watch_cb(int opcode, uint64_t ver, bufferlist& bl)
     cache.remove(name);
     break;
   default:
-    ldout(T::cct, 0) << "WARNING: got unknown notification op: " << info.op << dendl;
+    mydout(0) << "WARNING: got unknown notification op: " << info.op << dendl;
     return -EINVAL;
   }
 
